package de.unikoblenz.fgbks.base.utils.boundary;

import static de.unikoblenz.fgbks.base.utils.boundary.BoundType.INCLUSIVE;

import de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheck;
import de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckEqual;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckInContact;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckIsSubsumes;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckNotInContact;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckOverlap;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckSubsumes;
import java.util.HashMap;
import java.util.Objects;
import org.apache.commons.lang3.Validate;

public abstract class AbstractGrowingBoundary<T extends Comparable<? super T>>
    extends AbstractBoundary<T> {

  protected T lowerBound;
  protected BoundType lowerBoundType;
  protected T upperBound;
  protected BoundType upperBoundType;

  public T getLowerBound() {
    return lowerBound;
  }

  public T getUpperBound() {
    return upperBound;
  }

  public BoundType getLowerBoundType() {
    return lowerBoundType;
  }

  public BoundType getUpperBoundType() {
    return upperBoundType;
  }

  public AbstractGrowingBoundary(String text, Class<T> type) {
    super(text, type);
  }

  @Override
  protected void validate() {
    Validate.notNull(lowerBound);
    Validate.notNull(upperBound);
    if (lowerBound.equals(getMinValue())) {
      lowerBoundType = INCLUSIVE;
    }
    if (upperBound.equals(getMaxValue())) {
      upperBoundType = INCLUSIVE;
    }
    Validate.notNull(lowerBoundType);
    Validate.notNull(upperBoundType);
    if (lowerBound.compareTo(upperBound) > 0) {
      throw new IllegalArgumentException(
          String.format("Lower bound %s is greater than upper bound %s.", lowerBound, upperBound));
    }
    // if equal -> both must be INCLUSIVE!
    if (upperBound.equals(lowerBound)
        && (lowerBoundType != INCLUSIVE || upperBoundType != INCLUSIVE)) {
      throw new IllegalArgumentException("If bounds are equal -> both bounds must be INCLUSIVE!");
    }
  }

  @Override
  protected HashMap<BoundaryCheckType, BoundaryCheck> getChecker() {
    HashMap<BoundaryCheckType, BoundaryCheck> map = new HashMap<>();
    map.put(BoundaryCheckType.IS_EQUAL, CheckEqual.getInstance());
    map.put(BoundaryCheckType.IS_IN_CONTACT, CheckInContact.getInstance());
    map.put(BoundaryCheckType.IS_NOT_IN_CONTACT, CheckNotInContact.getInstance());
    map.put(BoundaryCheckType.IS_OVERLAPPING, CheckOverlap.getInstance());
    map.put(BoundaryCheckType.IS_SUBSUMPTION, CheckIsSubsumes.getInstance());
    map.put(BoundaryCheckType.SUBSUMES, CheckSubsumes.getInstance());
    return map;
  }

  @Override
  public int hashCode() {
    return Objects.hash(lowerBound, lowerBoundType, upperBound, upperBoundType);
  }

  @Override
  public int compareTo(Boundary o) {
    if (!(o instanceof AbstractGrowingBoundary)) {
      return 1;
    }
    int comLower = compareToLowerBound((AbstractGrowingBoundary) o);
    return comLower == 0 ? compareToUpperBound((AbstractGrowingBoundary) o) : comLower;
  }

  /**
   * Compare the lower bounds.
   *
   * @param o other boundary
   * @return -1 if current lower bound is lower, 0 if equal, 1 if higher
   */
  public int compareToLowerBound(AbstractGrowingBoundary o) {
    if (lowerBound.equals(o.lowerBound)) {
      if (lowerBoundType.equals(o.lowerBoundType)) {
        return 0;
      }
      return lowerBoundType == INCLUSIVE ? -1 : 1;
    } else if (lowerBound.equals(getMinValue())) {
      return -1;
    } else if (o.lowerBound.equals(getMinValue())) {
      return 1;
    }
    return lowerBound.compareTo((T) o.lowerBound);
  }

  /**
   * Compare the upper bounds.
   *
   * @param o other boundary
   * @return -1 if current upper bound is lower, 0 if equal, 1 if higher
   */
  public int compareToUpperBound(AbstractGrowingBoundary o) {
    if (upperBound.equals(o.upperBound)) {
      if (upperBoundType.equals(o.upperBoundType)) {
        return 0;
      }
      return upperBoundType == INCLUSIVE ? 1 : -1;
    } else if (upperBound.equals(getMaxValue())) {
      return 1;
    } else if (o.upperBound.equals(o.getMaxValue())) {
      return -1;
    }
    return upperBound.compareTo((T) o.upperBound);
  }

  protected abstract T getMinValue();

  protected abstract T getMaxValue();

  @Override
  protected String getParsedText() {
    return "";
  }
}
